import typer
from terminaltables import AsciiTable
import inquirer

from savvihub.savvihub_cli.inquirer import get_choices, parse_id
from savvihub.common.savvihub import SavviHubClient
from savvihub.common.constants import DEFAULT_CONFIG_PATH, CUR_DIR, INQUIRER_NAME_IMAGE, INQUIRER_NAME_RESOURCE, \
    INQUIRER_NAME_COMMAND, WEB_HOST, INQUIRER_NAME_DATASET, INQUIRER_NAME_DATASET_REF, \
    INQUIRER_NAME_DATASET_MOUNT_PATH, DEFAULT_SAVVIHUBFILE_YAML
from savvihub.savvihub_cli.errors import get_error_message
from savvihub.common.utils import *
from savvihub.savvihub_cli.yml_loader import ExperimentYmlLoader

experiment_app = typer.Typer()


@experiment_app.callback()
def main():
    """
    Perform your experiment with Savvihub
    """
    return


@experiment_app.command()
def init(
        slug: str = typer.Argument(..., help="Type workspace/project as an argument"),
        file: str = typer.Option(DEFAULT_SAVVIHUBFILE_YAML, "--file", "-f"),
):
    """
    Initialize a new experiment configuration file with workspace/project
    """
    # Remove old config file if exists
    config_file_path = os.path.join(CUR_DIR, file)
    remove_file(config_file_path)

    workspace, project = slug.split("/")
    data = {
        'workspace': workspace,
        'project': project,
    }

    make_file(config_file_path)
    yml_loader = ExperimentYmlLoader(config_file_path)
    yml_loader.write(data)

    typer.echo(f"Experiment config successfully made in {config_file_path}")


@experiment_app.command()
def data_mount(
        file: str = typer.Option(DEFAULT_SAVVIHUBFILE_YAML, "--file", "-f"),
):
    """
    Mount data to experiment
    """
    config_file_path = os.path.join(CUR_DIR, file)
    if not os.path.exists(config_file_path):
        raise Exception('Initialize experiment with this command: $ savvi experiment init SLUG')

    token = get_token_from_config(DEFAULT_CONFIG_PATH)
    client = SavviHubClient(token=token)

    yml_loader = ExperimentYmlLoader(file)

    questions = [
        inquirer.List(
            INQUIRER_NAME_DATASET,
            message="Please choose a dataset",
            choices=get_choices(client, 'dataset', yml_loader),
        ),
        inquirer.Text(
            INQUIRER_NAME_DATASET_REF,
            message="Dataset ref",
            default=yml_loader.data.get('dataset_mount_ref', 'latest'),
        ),
        inquirer.Text(
            INQUIRER_NAME_DATASET_MOUNT_PATH,
            message="Dataset mount path",
            default=yml_loader.data.get('dataset_mount_path', 'input'),
        ),
    ]

    answers = inquirer.prompt(questions)
    dataset_mount_id = parse_id(answers.get(INQUIRER_NAME_DATASET))

    yml_loader.update_and_write({
        'data_mount_infos': {
            'id': dataset_mount_id,
            'ref': answers.get(INQUIRER_NAME_DATASET_REF),
            'mount_path': answers.get(INQUIRER_NAME_DATASET_MOUNT_PATH),
        }
    })


@experiment_app.command()
def list(
        file: str = typer.Option(DEFAULT_SAVVIHUBFILE_YAML, "--file", "-f"),
):
    """
    List of experiments
    """
    token = get_token_from_config(DEFAULT_CONFIG_PATH)
    client = SavviHubClient(token=token)
    yml_loader = ExperimentYmlLoader(file)

    workspace = yml_loader.data.get('workspace')
    project = yml_loader.data.get('project')

    resp = client.experiment_list(workspace, project, raise_error=True)
    experiments = resp.json().get('results')
    table = AsciiTable([
        ['Number', 'Status', 'Message'],
        *[[e.get('number'), e.get('status'), e.get('message')] for e in experiments],
    ])
    typer.echo(table.table)


@experiment_app.command()
def log(
        file: str = typer.Option(DEFAULT_SAVVIHUBFILE_YAML, "--file", "-f"),
        experiment_number: int = typer.Argument(...),
):
    """
    View experiment logs
    """
    token = get_token_from_config(DEFAULT_CONFIG_PATH)
    client = SavviHubClient(token=token)
    yml_loader = ExperimentYmlLoader(file)

    workspace = yml_loader.data.get('workspace')
    project = yml_loader.data.get('project')

    resp = client.experiment_log(workspace, project, experiment_number, raise_error=True)
    print(resp.text)


@experiment_app.command()
def run(
        file: str = typer.Option(DEFAULT_SAVVIHUBFILE_YAML, "--file", "-f"),
):
    """
    Run an experiment in Savvihub
    """
    if not is_committed():
        raise Exception('You should commit diffs before run an experiment!')

    config_file_path = os.path.join(CUR_DIR, file)
    if not os.path.exists(config_file_path):
        raise Exception('Initialize experiment with this command: $ savvi experiment init SLUG')

    token = get_token_from_config(DEFAULT_CONFIG_PATH)
    client = SavviHubClient(token=token)

    yml_loader = ExperimentYmlLoader(file)

    if not yml_loader.is_ready_to_run():
        raise Exception('You should mount data first with this command: $ savvi experiment data-mount')

    questions = [
        inquirer.List(
            INQUIRER_NAME_IMAGE,
            message="Please choose a kernel image",
            choices=get_choices(client, 'image', yml_loader),
        ),
        inquirer.List(
            INQUIRER_NAME_RESOURCE,
            message="Please choose a kernel resource",
            choices=get_choices(client, 'resource', yml_loader),
        ),
        inquirer.Text(
            INQUIRER_NAME_COMMAND,
            message="Start command",
            default="python main.py",
        )
    ]

    answers = inquirer.prompt(questions)
    image_id = parse_id(answers.get(INQUIRER_NAME_IMAGE))
    resource_spec_id = parse_id(answers.get(INQUIRER_NAME_RESOURCE))

    yml_loader.update_and_write({
        'image_id': image_id,
        'resource_spec_id': resource_spec_id,
        'start_command': answers.get(INQUIRER_NAME_COMMAND),
    })

    res = client.experiment_create(
        workspace=yml_loader.data.get('workspace'),
        project=yml_loader.data.get('project'),
        image_id=int(yml_loader.data.get('image_id')),
        resource_spec_id=int(yml_loader.data.get('resource_spec_id')),
        branch=get_git_revision_hash(),
        dataset_mount_infos=[{
            'id': int(dataset_mount_info.get('id')),
            'ref': dataset_mount_info.get('ref'),
            'mount_path': dataset_mount_info.get('mount_path'),
        } for dataset_mount_info in yml_loader.data.get('data_mount_infos')],
        start_command=yml_loader.data.get('start_command'),
    )

    res_data = res.json()
    if res.status_code == 400:
        typer.echo(get_error_message(res_data))
        return
    res.raise_for_status()

    experiment_number = res_data.get('number')
    typer.echo(f"Experiment {experiment_number} is running. Check the experiment status at below link")
    typer.echo(f"{WEB_HOST}/{yml_loader.data.get('workspace')}/{yml_loader.data.get('project')}/"
               f"experiments/{experiment_number}")
    return


if __name__ == "__main__":
    experiment_app()



import os

import requests

from savvihub.common.config import config
from savvihub.common.experiment import Experiment
from savvihub.common.savvihub import SavviHubClient


experiment = None
token = None


def get_token():
    global token
    if not token:
        token = os.getenv('EXPERIMENT_ACCESS_TOKEN')
    return token


def get_experiment():
    global experiment

    if not experiment:
        token = get_token()
        if not token:
            print("EXPERIMENT_ACCESS_TOKEN is not provided as an environment variable")

        client = SavviHubClient(token=token)

        try:
            savvihub_experiment = client.experiment_read(raise_error=True).json()
        except requests.exceptions.HTTPError as e:
            print(f"Http Error: {e}")
            return
        except requests.exceptions.ConnectionError as e:
            print(f"Connection Error:{e}")
            return
        except requests.exceptions.Timeout as e:
            print(f"Timeout Error: {e}")
            return
        except requests.exceptions.TooManyRedirects as e:
            print(f"Too many Redirects Error: {e}")
            return
        except requests.exceptions.RequestException as e:
            raise SystemExit(e)

        config.experiment_id = savvihub_experiment.get("id")
        experiment = Experiment.from_given_or_global_config(config)

    return experiment



import inspect
import json
import os
import shutil
import subprocess

import zlib
from pathlib import Path

import typeguard
import yaml


def make_file(path):
    if not os.path.exists(path):
        with open(path, 'w'):
            print(" [*] Make a file : {}".format(path))
            pass


def make_dir(path):
    if not os.path.exists(path):
        print(" [*] Make directories : {}".format(path))
        os.makedirs(path)


def remove_file(path):
    if os.path.exists(path):
        print(" [*] Removed: {}".format(path))
        os.remove(path)


def remove_dir(path):
    if os.path.exists(path):
        print(" [*] Removed: {}".format(path))
        shutil.rmtree(path)


def write_json(path, data):
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(data, f, indent=4, sort_keys=True, ensure_ascii=False)


def load_json(path):
    with open(path) as f:
        data = json.loads(f.read())
    return data


class AnnotatedObject:
    def __init__(self, d: dict):
        resolved_types = {}
        for cls in self.__class__.mro():
            annotations = getattr(cls, '__annotations__', None)
            if annotations is None:
                continue

            for k, type_ in annotations.items():
                if k in resolved_types:
                    continue

                v = d.get(k, None)
                resolved_types[k] = type_
                typeguard.check_type(k, v, type_)

                if inspect.isclass(type_) and issubclass(type_, AnnotatedObject):
                    v = type_(v)
                setattr(self, k, v)
        self.d = d

    def __str__(self):
        return f'[{self.__class__.__name__}] {self.d}'

    @property
    def dict(self):
        return self.d


def get_token_from_config(path):
    try:
        with open(path) as f:
            documents = yaml.full_load(f)
            if not documents:
                raise Exception('Nothing in config file')
            for item, doc in documents.items():
                if item != 'token':
                    raise Exception('No token item')
                elif not doc:
                    raise Exception('Token is not set! Run $savvi init')
                else:
                    return doc
    except EnvironmentError:
        print('Error: Config file not found')


def get_json_headers_with_token(token):
    try:
        authentication = 'token ' + token
    except TypeError:
        print('Token is not set!')
        return

    headers = {
        'Content-Type': 'application/json',
        'Authentication': authentication
    }

    return headers


def calculate_crc32c(filename):
    with open(filename, 'rb') as fh:
        h = 0
        while True:
            s = fh.read(65536)
            if not s:
                break
            h = zlib.crc32(s, h)
        return "%X" % (h & 0xFFFFFFFF)


def read_in_chunks(filename, blocksize=65535, chunks=-1):
    """Lazy function (generator) to read a file piece by piece."""
    with open(filename, 'rb') as f:
        while chunks:
            data = f.read(blocksize)
            if not data:
                break
            yield data
            chunks -= 1


def get_active_branch_name():
    head_dir = Path(".") / ".git" / "HEAD"
    with head_dir.open("r") as f:
        content = f.read().splitlines()

    for line in content:
        if line[0:4] == "ref:":
            return line.partition("refs/heads/")[2]


def get_git_revision_hash():
    return subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode("utf-8")


def is_committed():
    from git import Repo
    repo = Repo(search_parent_directories=True)
    changed_files = [item.a_path for item in repo.index.diff(None)]
    for file in changed_files:
        if file == 'savvihubfile.yml' or file == '.gitignore':
            continue
        else:
            return False
    return True


from datetime import datetime

from savvihub.common.constants import DEFAULT_CONFIG_PATH
from savvihub.common.utils import get_token_from_config
from savvihub.common.savvihub import SavviHubClient


class History:
    def __init__(self, experiment):
        self.experiment = experiment
        self.rows = []
        self._step_index = 0

    def update(self, step, row):
        """
        Update row in history
        """
        for k, v in row.items():
            if type(v) != str:
                row[k] = str(v)

        row["step"] = str(step)
        row["created_dt"] = datetime.utcnow().isoformat()
        self.request(row)
        self.rows.append(row)

    def request(self, row):
        token = get_token_from_config(DEFAULT_CONFIG_PATH)
        client = SavviHubClient(token=token)
        _ = client.experiment_progress_update(self.experiment.experiment_id, row)



from savvihub.common.store import get_experiment
from savvihub.common.utils import get_token_from_config


def log(step, row=None):
    """
    step: a step for each iteration (required)
    row: a dictionary to log
    """
    experiment = get_experiment()
    experiment.log(step, row)



import requests
import typer

from savvihub.savvihub_cli.artifact import artifact_app
from savvihub.common.config import config
from savvihub.common.constants import API_HOST, DEFAULT_SAVVI_DIR, DEFAULT_CONFIG_PATH
from savvihub.savvihub_cli.dataset import dataset_app
from savvihub.savvihub_cli.experiment import experiment_app
from savvihub.common.utils import *

app = typer.Typer()
app.add_typer(experiment_app, name="experiment")
app.add_typer(dataset_app, name="dataset")
app.add_typer(artifact_app, name="artifact")
__version__ = '0.0.9'


def version_callback(value: bool):
    if value:
        typer.echo(f"Savvihub CLI Version: {__version__}")
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=True),
):
    """
    Savvihub command line interface
    """

    return


@app.command()
def ping():
    """
    Ping to server
    """
    res = requests.get(API_HOST + '/v1/api/ping/')
    typer.echo(f"Response code: {res.status_code}, Response text: {res.text}")


@app.command()
def init(
    token: str = typer.Option(config.token, "-t", "--token")
):
    """
    Initialize with an access token issued from SavviHub
    """
    if token is None:
        token = typer.prompt('token')

    data = {
        'token': token
    }

    make_dir(DEFAULT_SAVVI_DIR)
    with open(DEFAULT_CONFIG_PATH, 'w') as f:
        yaml.dump(data, f, default_flow_style=False)

    typer.echo(f"Token successfully saved in {DEFAULT_CONFIG_PATH}")
